#define __SP_CHARS_C__

/*
 * SPChars - parent class for text objects
 *
 * Author:
 *   Lauris Kaplinski <lauris@kaplinski.com>
 *
 * Copyright (C) 1999-2002 Lauris Kaplinski
 * Copyright (C) 2000-2001 Ximian, Inc.
 *
 * Released under GNU GPL, read the file 'COPYING' for more information
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <string.h>

//#include <libart_lgpl/art_misc.h>

#include <libnr/nr-macros.h>
#include <libnr/nr-matrix.h>
#include <libnr/nr-matrix-ops.h>
#include <libnr/nr-pixblock.h>
#include <libnrtype/font-instance.h>

#include "macros.h"
#include "display/nr-arena-glyphs.h"
#include "print.h"
#include "style.h"
#include "sp-chars.h"

static void sp_chars_class_init (SPCharsClass *klass);
static void sp_chars_init (SPChars *chars);

static void sp_chars_release (SPObject *object);
static void sp_chars_modified (SPObject *object, guint flags);

static void sp_chars_bbox(SPItem *item, NRRect *bbox, NR::Matrix const &transform, unsigned const flags);
static NRArenaItem *sp_chars_show (SPItem *item, NRArena *arena, unsigned int key, unsigned int flags);

static SPItemClass *parent_class;

GType
sp_chars_get_type (void)
{
	static GType type = 0;
	if (!type) {
		GTypeInfo info = {
			sizeof (SPCharsClass),
			NULL,	/* base_init */
			NULL,	/* base_finalize */
			(GClassInitFunc) sp_chars_class_init,
			NULL,	/* class_finalize */
			NULL,	/* class_data */
			sizeof (SPChars),
			16,	/* n_preallocs */
			(GInstanceInitFunc) sp_chars_init,
			NULL,	/* value_table */
		};
		type = g_type_register_static (SP_TYPE_ITEM, "SPChars", &info, (GTypeFlags)0);
	}
	return type;
}

static void
sp_chars_class_init (SPCharsClass *klass)
{
	GObjectClass *object_class;
	SPObjectClass *sp_object_class;
	SPItemClass *item_class;

	object_class = (GObjectClass *) klass;
	sp_object_class = (SPObjectClass *) klass;
	item_class = (SPItemClass *) klass;

	parent_class = (SPItemClass *)g_type_class_ref (SP_TYPE_ITEM);

	sp_object_class->release = sp_chars_release;
	sp_object_class->modified = sp_chars_modified;

	item_class->bbox = sp_chars_bbox;
	item_class->show = sp_chars_show;
}

static void
sp_chars_init (SPChars *chars)
{
	chars->elements = NULL;

	chars->paintbox.x0 = chars->paintbox.y0 = 0.0;
	chars->paintbox.x1 = chars->paintbox.y1 = 1.0;
}

static void
sp_chars_release (SPObject *object)
{
	SPChars *chars;

	chars = SP_CHARS (object);

	while (chars->elements) {
		SPCharElement *el;
		el = chars->elements;
		chars->elements = el->next;
		//printf("sp_char_release ");
		if ( el->font ) el->font->Unref();
		g_free (el);
	}

	if (((SPObjectClass *) parent_class)->release)
		((SPObjectClass *) parent_class)->release (object);
}

static void
sp_chars_modified (SPObject *object, unsigned int flags)
{
	SPChars *chars;

	chars = SP_CHARS (object);

	if (((SPObjectClass *) (parent_class))->modified)
		((SPObjectClass *) (parent_class))->modified (object, flags);

	if (flags & SP_OBJECT_STYLE_MODIFIED_FLAG) {
		SPItemView *v;
		for (v = SP_ITEM (chars)->display; v != NULL; v = v->next) {
			nr_arena_glyphs_group_set_style (NR_ARENA_GLYPHS_GROUP (v->arenaitem), object->style);
		}
	}
}

static void
sp_chars_bbox(SPItem *item, NRRect *bbox, NR::Matrix const &transform, unsigned const flags)
{
	SPChars * const chars = SP_CHARS(item);

	for (SPCharElement *el = chars->elements; el != NULL; el = el->next) {
		NRBPath bpath;
		if ( el->font ) bpath.path=(NArtBpath*)el->font->ArtBPath(el->glyph); else bpath.path=NULL;
		if ( bpath.path ) {
			NR::Matrix const a( el->transform * transform );
			nr_path_matrix_bbox_union(&bpath, a, bbox);
		}
	}
}

static NRArenaItem *
sp_chars_show (SPItem *item, NRArena *arena, unsigned int key, unsigned int flags)
{
	SPChars *chars;
	NRArenaItem *arenaitem;
	SPCharElement *el;

	chars = SP_CHARS (item);

	arenaitem = NRArenaGlyphsGroup::create(arena);

	nr_arena_glyphs_group_set_style (NR_ARENA_GLYPHS_GROUP (arenaitem), SP_OBJECT_STYLE (item));

	for (el = chars->elements; el != NULL; el = el->next) {
		nr_arena_glyphs_group_add_component (NR_ARENA_GLYPHS_GROUP (arenaitem), el->font, el->glyph, &el->transform);
	}

	nr_arena_glyphs_group_set_paintbox (NR_ARENA_GLYPHS_GROUP (arenaitem), &chars->paintbox);

	return arenaitem;
}

void
sp_chars_clear (SPChars *chars)
{
	SPItem *item;
	SPItemView *v;

	item = SP_ITEM (chars);

	while (chars->elements) {
		SPCharElement *el;
		el = chars->elements;
		chars->elements = el->next;
		//printf("sp_char_clear ");
		if ( el->font ) el->font->Unref();
		g_free (el);
	}

	for (v = item->display; v != NULL; v = v->next) {
		nr_arena_glyphs_group_clear (NR_ARENA_GLYPHS_GROUP (v->arenaitem));
	}
}

void sp_chars_add_element(SPChars *chars, guint glyph, font_instance *font, NR::Matrix const &transform)
{
	NRMatrix const ntransform(transform);
	sp_chars_add_element(chars, glyph, font, &ntransform);
}

void
sp_chars_add_element (SPChars *chars, guint glyph, font_instance *font, const NRMatrix *transform)
{
	SPItem *item;
	SPItemView *v;
	SPCharElement * el;

	item = SP_ITEM (chars);

	el = g_new (SPCharElement, 1);

	el->glyph = glyph;
	el->font = font;
	//printf("sp_char_add_elem ");
	if ( font ) font->Ref();

	el->transform = *transform;

	el->next = chars->elements;
	chars->elements = el;

	for (v = item->display; v != NULL; v = v->next) {
		nr_arena_glyphs_group_add_component (NR_ARENA_GLYPHS_GROUP (v->arenaitem), el->font, el->glyph, &el->transform);
	}
}

SPCurve *
sp_chars_normalized_bpath (SPChars *chars)
{
	GSList *cc = NULL;
	for (SPCharElement *el = chars->elements; el != NULL; el = el->next) {
		NRBPath bp;
		if ( el->font ) bp.path=(NArtBpath*)el->font->ArtBPath(el->glyph); else bp.path=NULL;
		if ( bp.path ) {
			NArtBpath *abp = nr_artpath_affine (bp.path, el->transform);
			SPCurve *c = sp_curve_new_from_bpath (abp);
			if (c)
				cc = g_slist_prepend (cc, c);
		}
	}

	cc = g_slist_reverse (cc);

	SPCurve *curve;
	if ( cc ) {
		curve = sp_curve_concat (cc);
	} else {
		curve = sp_curve_new();
	}
  
	while (cc) {
		/* fixme: This is dangerous, as we are mixing art_alloc and g_new */
		sp_curve_unref ((SPCurve *) cc->data);
		cc = g_slist_remove (cc, cc->data);
	}

	return curve;
}

/* This is completely unrelated to SPItem::print */

static void
sp_chars_print_bpath (SPPrintContext *ctx, const NRBPath *bpath, const SPStyle *style, const NRMatrix *ctm,
		      const NRRect *pbox, const NRRect *dbox, const NRRect *bbox)
{
	if (style->fill.type != SP_PAINT_TYPE_NONE) {
		sp_print_fill (ctx, bpath, ctm, style, pbox, dbox, bbox);
	}

	if (style->stroke.type != SP_PAINT_TYPE_NONE) {
		sp_print_stroke (ctx, bpath, ctm, style, pbox, dbox, bbox);
	}
}

/*
 * pbox is bbox for paint server (user coordinates)
 * dbox is the whole display area
 * bbox is item bbox on desktop
 */

void
sp_chars_do_print (SPChars *chars, SPPrintContext *ctx, const NRMatrix *ctm, const NRRect *pbox, const NRRect *dbox, const NRRect *bbox)
{
	SPCharElement *el;

	for (el = chars->elements; el != NULL; el = el->next) {
		NRBPath bpath;
		if ( el->font ) bpath.path=(NArtBpath*)el->font->ArtBPath(el->glyph); else bpath.path=NULL;
		if ( bpath.path ) {
			NRBPath abp;
			abp.path = nr_artpath_affine (bpath.path, el->transform);
			sp_chars_print_bpath (ctx, &abp, SP_OBJECT_STYLE (chars), ctm, pbox, dbox, bbox);
			nr_free (abp.path);
		}
	}
}

void
sp_chars_set_paintbox (SPChars *chars, NRRect *paintbox)
{
	SPItemView *v;

	chars->paintbox.x0 = paintbox->x0;
	chars->paintbox.y0 = paintbox->y0;
	chars->paintbox.x1 = paintbox->x1;
	chars->paintbox.y1 = paintbox->y1;

	for (v = SP_ITEM (chars)->display; v != NULL; v = v->next) {
		nr_arena_glyphs_group_set_paintbox (NR_ARENA_GLYPHS_GROUP (v->arenaitem), &chars->paintbox);
	}
}

